from __future__ import annotations

import base64
import struct
import zlib
from functools import reduce
from operator import mul
from typing import Any, Callable

from pysll.exceptions import UnknownTokenError
from pysll.feature_flags import FeatureFlags


def decode(compressed: str) -> str:
    """The encoded data in a compressed field is a zlib compression of a base64
    encoded string prepended with `1.` that serializes a MM expression.

    See
    https://mathematica.stackexchange.com/questions/104660/what-algorithm-do-the-compress-and-uncompress-functions-use
    for more info
    """
    return _decode(zlib.decompress(base64.b64decode(compressed[2:])))


def _decode(data: bytes) -> str:
    """Decodes the MM binary object representation into a MM expression.

    See
    https://mathematica.stackexchange.com/questions/104660/what-algorithm-do-the-compress-and-uncompress-functions-use
    for more info
    """
    expressions: list[str] = []
    while data:
        expression, data = decode_token(data)
        expressions.append(expression)

    return "".join(map(str, expressions))


def decode_token(data: bytes) -> tuple[Any, bytes]:
    """Parse a single token from the binary object representation."""

    # First, check if we're seeing the header
    if data[:4] == b"!boR":
        # Just strip the header
        return "", data[4:]

    if not data:
        return "", b""

    # If not, then check if we can read the leading character to determine the token type
    token, data = data[0], data[1:]
    match token:
        case 105:  # b"i"
            return _decode_integer(data)
        case 114:  # b"r"
            return _decode_real(data)
        case 83 | 115 | 73:  # b"S", b"s", b"I"
            # Strings, symbols, large integers
            return _decode_string(data)
        case 102:  # b"f"
            return _decode_expression(data)
        case 101:  # b"e"
            return _decode_tensor(data)
        case 82:  # b"R"
            # this should be parsable as a string which produces something like '123.123`1.12'
            # we will get the numeric string, and split out the first value, convert it to
            # a float and return. we will ignore the precision info b/c python doesn't care
            value, rest = _decode_string(data)
            if not isinstance(value, str):
                raise ValueError(f"Expected an arbitrary precision real string, and got {type(value)}")
            real, _ = value.split("`")

            return float(real), rest
        case _:
            raise UnknownTokenError(token, data)


def _decode_integer(data: bytes):
    return struct.unpack("<i", data[:4])[0], data[4:]


def _decode_real(data: bytes):
    return struct.unpack("<d", data[:8])[0], data[8:]


def _decode_string(data: bytes):
    # the first four bytes hold the size of the string
    size, data = _decode_integer(data)
    return data[:size].decode("utf-8"), data[size:]


def _decode_expression(data: bytes):
    # Integer tells us the expression size
    size, data = _decode_integer(data)

    # Next, we got the symbol head
    (
        head,
        data,
    ) = decode_token(data)

    # Finally, arguments.
    arguments = []
    for _ in range(size):
        if not data:
            break

        argument, data = decode_token(data)
        arguments.append(argument)

    return f"{head}[{', '.join(map(str, arguments))}]", data


def _decode_tensor(
    data: bytes,
    decoder: Callable[[bytes], tuple[Any, bytes]] | None = None,
):
    dimensions, data = _decode_integer(data)

    shapes = []
    for _ in range(dimensions):
        size, data = _decode_integer(data)
        shapes.append(size)

    if FeatureFlags.from_environment().use_numpy:
        import numpy as np  # need pip install pysll[numpy]

        n = reduce(mul, shapes) * 8
        return np.frombuffer(data[:n], dtype=np.float64).reshape(*shapes).tolist(), data[n:]

    return _reshape(tuple(shapes), data, decoder or _decode_real)


def _reshape(
    shapes: tuple[int, ...],
    data: bytes,
    decoder: Callable[[bytes], tuple[Any, bytes]],
) -> tuple[list, bytes]:
    if not shapes:
        return decoder(data)

    tensor = []
    head, *tail = shapes
    for _ in range(head):
        vector, data = _reshape(tuple(tail), data, decoder)
        tensor.append(vector)

    return tensor, data



from __future__ import annotations

import base64
import functools
import json
import os
import pathlib
import tempfile
import time
from dataclasses import asdict, dataclass
from http import HTTPStatus
from itertools import count
from typing import Any, Generator, List, Optional, TypeVar, Union

import backoff
import requests
import rollbar
import xmltodict

from .exceptions import (
    ConstellationException,
    ConstellationInvalidMethodException,
    ConstellationInvalidResponseException,
    ConstellationMissingFieldException,
    ConstellationNotLoggedInException,
    ConstellationObjectDoesNotExistException,
)
from .models import (
    BlobRef,
    CloudFileArgs,
    ConstellationFieldParser,
    Function,
    ListableKind,
    ListableType,
    Model,
    Object,
    ResultPayload,
    deserialize_item,
)
from .parsers import (
    parse_download_field_string,
    parse_field_values_from_response,
    parse_results_from_response,
)
from .utils import create_two_way_link, extend, md5_hash_file, parts_to_xml, tmap

LK = TypeVar("LK", bound=ListableKind)


class Constellation:
    _UPLOAD_PART_SIZE = 10_000_000  # 10 MB
    _MULTIPART_CLOUDFILE_THRESHOLD = 100_000_000  # 100 MB

    def __init__(self, auth_token="", host="", notebook_id=None):
        self._host = host or "https://constellation.emeraldcloudlab.com"
        self._auth_token = auth_token

        self._temp_directory = None

        # try to set the notebook ID to the default if not already set
        if not notebook_id and auth_token != "":
            self._set_default_notebook_id()
        else:
            self._notebook_id = notebook_id

        # set the sleep time for getting API results here
        self._sleep_time = 0.3
        self._max_sleep_time = 600

    # Retry decorator for requests failures (like timeout or connection error)
    _retry_requests_decorator = backoff.on_exception(
        wait_gen=backoff.expo,
        exception=(requests.exceptions.Timeout, requests.exceptions.ConnectionError),
        max_tries=5,
    )

    # Configuration related functions
    def host(self):
        """Return the url of the instance constellation that the client is
        connecting to."""
        return self._host

    def download_directory(self):
        """Returns where cloud file downloads are stored."""
        if self._temp_directory is None:
            self._temp_directory = tempfile.TemporaryDirectory()

        return self._temp_directory.name

    def _set_default_notebook_id(self):
        me = Object(self.me().get("Id"))
        default_notebook = self.download(me, "DefaultNotebook")
        if not default_notebook:
            # attempt to download from the first notebook of the financing team
            notebook_lists = self.download(me, "FinancingTeams[Notebooks]")
            default_notebook = notebook_lists[0][0]
        if isinstance(default_notebook, Object):
            self._notebook_id = default_notebook.id
        else:
            raise ValueError(f"Cannot find a default notebook for the user: {me}")

    # Identity requests
    @_retry_requests_decorator
    def login(self, username, password):
        """Return the auth token token for the supplied username and
        password."""
        auth_token = self._send_and_validate_request(
            "/ise/signintoken",
            {"username": username, "password": password},
            "AuthToken",
            requires_auth_token=False,
        )
        self.set_auth_token(auth_token)
        # set the notebook ID if not already set
        if self._notebook_id is None:
            self._set_default_notebook_id()
        return self.me()

    def is_logged_in(self):
        return self._auth_token != ""

    def set_auth_token(self, auth_token):
        """Set the auth token used by the constellation client."""
        self._auth_token = auth_token

    @_retry_requests_decorator
    def me(self):
        """Return the response from the /ise/me endpoint."""
        return self._send_and_validate_request("/ise/me", None, None, method="GET")

    # Searching objects
    def search(self, ecl_type, query, max_results: int | None = None):
        """Perform a simple search on a given type."""
        payload = {
            "queries": [
                {
                    "clauses": [
                        {
                            "Types": [ecl_type],
                            "Query": query,
                        }
                    ],
                    "SubTypes": True,
                    "SoftLimit": max_results,
                    "date": None,
                    "ignore_time": False,
                }
            ],
        }

        return self._send_and_validate_request(
            "/obj/search",
            payload,
            field_to_retrieve="Results",
        )

    # Downloading objects
    def download(
        self,
        objects,
        field_string: Union[str, List[str]] = "All",
        auto_download_cloud_files=False,
        byte_size: Optional[int] = None,
    ) -> Any:
        """Download the supplied fields off the supplied object or objects.

        Note that fields can traverse across links using []'s.  For
        example:

        download(self.me()["ID"], "FinancingTeams[Name]")

        will download the status field off the protocol link of the
        supplied object.
        """
        result = None
        collapse_fields = False
        collapse_objects = False

        # Normalize the object and fields so they're always lists,
        # even if they're passed as single things.  This allows to
        # handle things like download("obj1", "Name") and
        # download([obj1], ["Name"]) in the same interface
        if not isinstance(field_string, list):
            field_string = [field_string]
            collapse_fields = True

        if not isinstance(objects, list):
            objects = [objects]
            collapse_objects = True

        result = self._download_objects(objects, field_string, auto_download_cloud_files, byte_size=byte_size)

        # Unwind the normalization of fields and object ids into lists
        if collapse_fields and collapse_objects:
            return result[0][0]
        if collapse_fields:
            return [x[0] for x in result]
        if collapse_objects:
            return result[0]

        return result

    def _download_objects(self, objects, field_string, auto_download_cloud_files, byte_size: Optional[int] = None):
        """Download multiple objects in a single request."""

        # Build the request
        traversals = parse_download_field_string(field_string)
        obj_requests = []
        for obj in objects:
            obj_request = {"object": {"id": obj.id}, "limit": 0, "subfield_info": True}
            if traversals:
                obj_request["traversals"] = traversals
            obj_requests.append(obj_request)

        # Send the request to constellation
        responses = self._send_and_validate_request(
            "/obj/download",
            {"requests": obj_requests},
            "responses",
        )
        if len(responses) != len(objects):
            raise ConstellationInvalidResponseException(
                f"Expected exactly {len(objects)} response from /obj/download", responses
            )

        # Parse the json results into objects
        response_objects = {}
        field_parser = ConstellationFieldParser()
        for response in responses:
            (
                sub_response_objects,
                sub_response_field_summaries,
            ) = parse_field_values_from_response(response)
            response_objects.update(sub_response_objects)
            for ecl_type, field_summaries in sub_response_field_summaries.items():
                field_parser.add_summaries_for_type(field_summaries, ecl_type)

        # Map the objects into a formatted result based on the fields string
        parsed_results = parse_results_from_response(response_objects, objects, traversals, field_parser)

        # now we want to download all the blob refs all in one go, and parse them locally
        paths = self.download_blob_ref(field_parser._blob_refs_encountered)
        parsed_blobs = [field_parser.parse_local_file(p) for p in paths]
        # create a blob ref store from the connecting the blob ref file hashes to their parsed values
        blob_store = {blob.file_hash: value for blob, value in zip(field_parser._blob_refs_encountered, parsed_blobs)}

        # exchange the blobs in the results with their parsed values
        self._replace_blob_values(parsed_results, blob_store)

        # If requested, go ahead and download all the cloud files encountered
        if auto_download_cloud_files:
            self.download_cloud_file(field_parser.cloud_files_encountered(), byte_size=byte_size)

        # return the results
        return parsed_results

    def resolve_type(self, obj: LK) -> LK:
        """Queries Constellation for the object(s) type(s) and injects them
        into the input object(s).

        >>> obj = Object("id:qdkmxzGkAK0a")
        >>> obj
        Object["id:qdkmxzGkAK0a"]
        >>> client.resolve_type(obj)
        >>> obj
        Object[Sample, "id:qdkmxzGkAK0a"]
        """
        match obj:
            case Object() | Model():
                objects = [obj]
            case _:
                objects = list(obj)

        responses = self._send_and_validate_request(
            "/obj/download",
            {
                "requests": [
                    {
                        "object": {"id": obj.id},
                        "fields": ["ID"],
                    }
                    for obj in objects
                ]
            },
            "responses",
        )
        for object, response in zip(objects, responses):
            try:
                resolved_object = response["resolved_object"]
            except KeyError:
                raise ConstellationObjectDoesNotExistException(object.id)
            object.type = resolved_object["type"]

        return obj

    def _replace_blob_values(self, parsed_results: list, store: dict) -> None:
        """Replace the BlobRef values in the parsed results list with their
        associated python types that have been parsed from their downloaded
        file.

        Note that this function replaces the values in-line so the
        original parsed results list is mutated in the process.
        """
        for i in range(len(parsed_results)):
            pr = parsed_results[i]
            if isinstance(pr, list):
                self._replace_blob_values(pr, store)
            elif isinstance(pr, BlobRef):
                parsed_results[i] = store[pr.file_hash]
        return None

    # Downloading cloud files
    def download_cloud_file(self, cloud_files: list[Object] | Object, byte_size: Optional[int] = None):
        """Download the supplied cloud file to client.download_directory()"""
        # process singleton inputs into lists
        singleton_flag = False
        if not isinstance(cloud_files, list):
            cloud_files = [cloud_files]
            singleton_flag = True

        # download the s3 information in one go
        cf_info_list = self.download(cloud_files, ["FileName", "FileType", "CloudFile"])

        # unpack the results
        file_names = [
            cfi[0] if cfi[0] else f"EmeraldCloudFile_{cf.id.replace(':', '_')}"
            for cfi, cf in zip(cf_info_list, cloud_files)
        ]
        extensions = [cfi[1].replace('"', "") for cfi in cf_info_list]
        file_infos = [cfi[2] for cfi in cf_info_list]
        target_paths = [self._get_temp_file_path(fn, ext) for fn, ext in zip(file_names, extensions)]

        # prepare a function to thread over by enclosing the bytesize option into the main helper function
        # yielding a lambda that takes in 2 arguments, a file_info and a path
        def parallel_func(file_info: dict, path: str):
            return self._download_cloud_file(
                file_info["CloudFileId"], file_info["Bucket"], file_info["Key"], path, byte_size=byte_size
            )

        # parallel map over the s3 downloads
        target_paths = tmap(parallel_func, file_infos, target_paths)

        # update the cloud file objects with the path information
        for cf, tp in zip(cloud_files, target_paths):
            cf.local_path = tp

        # if a singleton file was requested, unpack the list into a singleton object
        if singleton_flag:
            target_paths = target_paths[0]
        return target_paths

    def _get_temp_file_path(self, filename, extension):
        """Get a temp file path as close to filename as possible."""
        index = 0
        target_path = os.path.join(self.download_directory(), f"{filename}.{extension}")
        while os.path.exists(target_path):
            target_path = os.path.join(self.download_directory(), f"{filename}_{index}.{extension}")
            index += 1

        return target_path

    def blobsign_download(self, cloud_file_id, cloud_file_bucket, cloud_file_key):
        """Return the url to download the supplied cloud file."""
        return self._send_and_validate_request(
            "/blobsign/download",
            {
                "CloudFileId": cloud_file_id,
                "Bucket": cloud_file_bucket,
                "Key": cloud_file_key,
            },
            "Url",
        )

    def _download_cloud_file(
        self, cloud_file_id, cloud_file_bucket, cloud_file_key, target_path, byte_size: Optional[int] = None
    ):
        """Download a cloud file based off the supplied identifiers.

        Default byte_size is None, in which case file will download in a
        single request. Note, byte ranges are inclusive, so getting
        0-1024, and then 1024-2048 would downlaod the 1024th byte twice.
        """

        url = self.blobsign_download(cloud_file_id, cloud_file_bucket, cloud_file_key)
        if not url:
            raise ConstellationMissingFieldException("When downloading cloud file", "url")

        # Now download the file from s3
        target_file = open(target_path, "wb")
        if byte_size is None:
            download_response = requests.get(url)
            target_file.write(download_response.content)
        else:
            assert byte_size > 0, "Expected `byte_size` value to be positive"
            chunks = lambda w: ((lower, lower + w - 1) for lower in count(0, step=w))  # noqa: E731
            for lower, upper in chunks(byte_size):
                download_response = requests.get(url, headers={"Range": f"bytes={lower}-{upper}"})
                if download_response.status_code != HTTPStatus.PARTIAL_CONTENT:
                    break
                target_file.write(download_response.content)

        return target_path

    def download_blob_ref(
        self, blob_refs: list[BlobRef] | BlobRef, target_path="", byte_size: Optional[int] = None
    ) -> list[str] | str:
        """Download a blob ref to the supplied target path."""
        singleton_flag = not isinstance(blob_refs, list)
        if singleton_flag:
            blob_refs = [blob_refs]
        target_paths = [target_path or self._get_temp_file_path(br.file_hash, "tmp") for br in blob_refs]

        cloud_file_buckets = [br.bucket for br in blob_refs]
        cloud_file_keys = [br.key() for br in blob_refs]
        # check the buckets
        if any(not bucket for bucket in cloud_file_buckets):
            raise ConstellationMissingFieldException("When downloading cloud file", "bucket")
        if any(not key for key in cloud_file_keys):
            raise ConstellationMissingFieldException("When downloading cloud file", "hash")

        parallel_fn = functools.partial(self._download_cloud_file, cloud_file_id="", byte_size=byte_size)
        target_paths = tmap(
            parallel_fn, cloud_file_bucket=cloud_file_buckets, cloud_file_key=cloud_file_keys, target_path=target_paths
        )
        # update the blob ref paths
        for br, tp in zip(blob_refs, target_paths):
            br.local_path = tp

        # unpack the singleton if necessary
        if singleton_flag:
            target_paths = target_paths[0]
        return target_paths

    def upload(self, object_type, object_id, new_field_values, allow_public_objects: bool = False):
        """Performs a very simple upload of a list of fields on a single
        object.

        Pass `None` for object_id to create a new object. By default, new objects will be linked to
        the requesting user's default notebook (or their financing team's default notebook). In order
        to allow for public objects to be created, the `allow_public_objects` flag need to be explicitly
        set to `True`.
        """
        request_body = {
            "object": {"id": object_id, "type": object_type},
            "fields": new_field_values,
        }
        if object_id is None:
            # If you have no object id because you're creating a new object,
            # you need to format it slightly differently
            request_body = {"type": object_type, "fields": new_field_values}
        responses = self._send_and_validate_request(
            "/obj/upload",
            {"requests": [request_body]},
            "responses",
            method="PUT",
            additional_headers={"X-ECL-AllowPublicObjects": "true"} if allow_public_objects else None,
        )

        if len(responses) != 1:
            raise ConstellationInvalidResponseException("Expected exactly 1 response from /obj/upload", responses)

        return responses[0]

    # Uploading cloud files
    def upload_cloud_file(self, file_path, notebook_id: str | None = None):
        """Uploads the given file to constellation."""
        md5 = md5_hash_file(file_path)
        extension = pathlib.Path(file_path).suffix
        if extension:
            extension = extension[1:]  # Remove the dot
        with open(file_path, "rb") as file:
            cloud_file_args = CloudFileArgs(
                file=file,
                name=pathlib.Path(file_path).stem,
                extension=extension,
                path=file_path,
                size=os.path.getsize(file_path),
                key=f"{md5.hexdigest()}.{extension}",
                content_md5=base64.b64encode(md5.digest()).decode(),
                content_disposition=f'attachment; filename="{file_path}"',  # noqa
            )

            method = (
                self._single_sign_and_upload_file_to_s3
                if cloud_file_args.size < self._MULTIPART_CLOUDFILE_THRESHOLD
                else self._multi_part_sign_and_upload_file_to_s3
            )

            return method(cloud_file_args, notebook_id)

    @_retry_requests_decorator
    def _single_sign_and_upload_file_to_s3(self, cloud_file_args, notebook_id: str | None = None):
        blobsign_response = self._blobsign("upload", cloud_file_args)
        headers = {
            "Content-Md5": cloud_file_args.content_md5,
            "Content-Disposition": cloud_file_args.content_disposition,
        }
        cloud_file_args.file.seek(0)
        s3_response = requests.put(
            url=blobsign_response["Url"],
            data=cloud_file_args.file,
            headers=headers,
        )
        if s3_response.status_code != HTTPStatus.OK:
            raise ConstellationException(
                f"Unexpected non-200 response from constellation endpoint {s3_response}: {s3_response.text}",
                s3_response.status_code,
            )
        return self._upload_cloud_file_to_constellation(blobsign_response["Key"], cloud_file_args, notebook_id)

    @_retry_requests_decorator
    def _upload_cloud_file_to_constellation(self, s3_key, cloud_file_args, notebook_id: str | None = None):
        cloud_field_values = {
            "$Type": "__JsonEmeraldCloudFile__",
            "Bucket": "emeraldsci-ecl-blobstore-stage",
            "CloudFileId": "None",
            "Key": s3_key,
        }
        fields = extend(
            {
                "FileName": cloud_file_args.name,
                "FileSize": f'Quantity[{cloud_file_args.size}, "bytes"]',
                "FileType": f'"{cloud_file_args.extension}"',
                "CloudFile": cloud_field_values,
            },
            (
                {
                    "Notebook": create_two_way_link(
                        linked_field="Objects", linked_id=notebook_id, linked_type="Object.LaboratoryNotebook"
                    )
                }
                if notebook_id is not None
                else {}
            ),
        )

        return self.upload("Object.EmeraldCloudFile", None, fields)

    def _blobsign(self, sign_type, cloud_file_args: CloudFileArgs, part=0, upload_id=""):
        """Return the url to upload the supplied cloud file."""
        sign_response = self._send_and_validate_request(
            "/blobsign/" + sign_type,
            {
                "Key": cloud_file_args.key,
                "ContentMD5": cloud_file_args.content_md5,
                "ContentDisposition": cloud_file_args.content_disposition,
                "Part": part,
                "UploadID": upload_id,
            },
            additional_headers={"ContentDisposition": cloud_file_args.content_disposition},
        )
        return sign_response

    def _multi_part_sign_and_upload_file_to_s3(self, cloud_file_args, notebook_id: str | None = None):
        upload_id, sharded_key = self._send_create_multi_part_upload_request(cloud_file_args)
        try:
            parts = self._upload_all_parts(upload_id, cloud_file_args)
            self._send_complete_multi_part_upload_request(upload_id, parts, cloud_file_args)
        except Exception as cloudFileUploadException:
            self._send_abort_multi_part_upload_request(upload_id, cloud_file_args)
            raise ConstellationException(
                "Unexpected exception while uploading cloud file", HTTPStatus.INTERNAL_SERVER_ERROR
            ) from cloudFileUploadException
        return self._upload_cloud_file_to_constellation(sharded_key, cloud_file_args, notebook_id)

    @_retry_requests_decorator
    def _send_create_multi_part_upload_request(self, cloud_file_args):
        blobsign_response = self._blobsign("init_multi_upload", cloud_file_args)
        s3_response = requests.post(
            url=blobsign_response["Url"],
        )
        if s3_response.status_code != HTTPStatus.OK:
            raise ConstellationException(
                f"Unexpected non-200 response from s3 endpoint {s3_response}: {s3_response.text}, "
                "failed to init multi-part upload",
                s3_response.status_code,
            )
        parsed_response = xmltodict.parse(s3_response.content)
        return (
            parsed_response["InitiateMultipartUploadResult"]["UploadId"],
            parsed_response["InitiateMultipartUploadResult"]["Key"],
        )

    @_retry_requests_decorator
    def _send_complete_multi_part_upload_request(self, upload_id, parts, cloud_file_args):
        blobsign_response = self._blobsign("complete_multi_upload", cloud_file_args, upload_id=upload_id)
        s3_response = requests.post(url=blobsign_response["Url"], data=parts_to_xml(parts))
        if s3_response.status_code != HTTPStatus.OK:
            raise ConstellationException(
                f"Unexpected non-200 response from s3 endpoint {s3_response}: {s3_response.text}, "
                "failed to complete multi-part upload",
                s3_response.status_code,
            )

    @_retry_requests_decorator
    def _send_abort_multi_part_upload_request(self, upload_id, cloud_file_args):
        blobsign_response = self._blobsign("abort_multi_upload", cloud_file_args)
        s3_response = requests.delete(url=blobsign_response["Url"])
        if s3_response.status_code != HTTPStatus.OK:
            rollbar.report_message(
                "Failed to abort multi part upload. "
                f"UploadID: {upload_id}, Key: {blobsign_response['Key']}, "
                f"S3 response: {s3_response} {s3_response.text}",
                "error",
            )
            raise ConstellationException(
                f"Unexpected non-200 response from s3 {s3_response}: {s3_response.text}, "
                "failed to abort multi part upload",
                s3_response.text,
            )

    def _upload_all_parts(self, upload_id, cloud_file_args):
        current_part_number = 1
        parts = []
        chunk = cloud_file_args.file.read(self._UPLOAD_PART_SIZE)
        while chunk != b"":
            part = self._upload_part(upload_id, current_part_number, chunk, cloud_file_args)
            parts.append(part)
            current_part_number += 1
            chunk = cloud_file_args.file.read(self._UPLOAD_PART_SIZE)
        return parts

    @_retry_requests_decorator
    def _upload_part(self, upload_id, part_number, data_chunk, cloud_file_args):
        blobsign_response = self._blobsign("upload_part", cloud_file_args, part=part_number, upload_id=upload_id)
        s3_response = requests.put(
            url=blobsign_response["Url"],
            data=data_chunk,
        )
        if s3_response.status_code != HTTPStatus.OK:
            raise ConstellationException(
                f"Unexpected non-200 response from s3 endpoint {s3_response}: {s3_response.text}, "
                "failed to upload part",
                s3_response.status_code,
            )
        return {"ETag": s3_response.headers["ETag"], "PartNumber": part_number}

    # Getting history of changes to an object
    def object_log(
        self,
        object_ids=None,
        types=None,
        start_date=None,
        end_date=None,
        max_results=1000,
    ):
        """Find changes on the supplied types after the supplied start_date."""
        payload = {
            "limit": max_results,
            "order": "dsc",
        }
        if object_ids is not None:
            payload["object"] = [{"id": x} for x in object_ids]
        if types is not None:
            payload["Types"] = types
        if start_date is not None:
            payload["startDate"] = start_date
        if end_date is not None:
            payload["endDate"] = end_date
        return self._send_and_validate_request(
            "/obj/objectlog",
            payload,
        )

    # Getting type information
    def get_type(self, type_name):
        """Retrieve all of the fields and their formats for a given type."""
        return self._send_and_validate_request(
            "/obj/type/{}".format(type_name),
            {},
            method="GET",
        )

    # Internal helper functions
    def _auth_header_from_token(self, auth_token):
        """Build a authorization header when an auth token is supplied."""
        return {"Authorization": "Bearer " + str(auth_token)}

    def _send_and_validate_request(
        self,
        path,
        data,
        field_to_retrieve=None,
        method="POST",
        requires_auth_token=True,
        additional_headers=None,
    ):
        """Send a request and validate a request to the given path."""
        if requires_auth_token and not self._auth_token:
            raise ConstellationNotLoggedInException()

        url = self._host + path
        if additional_headers is None:
            additional_headers = {}
        headers = {"Content-type": "application/json", **additional_headers}
        if self._auth_token:
            headers.update(self._auth_header_from_token(self._auth_token))
        if method == "POST":
            response = requests.post(
                url=url,
                data=json.dumps(data),
                headers=headers,
            )
        elif method == "PUT":
            response = requests.put(
                url=url,
                data=json.dumps(data),
                headers=headers,
            )
        elif method == "GET":
            response = requests.get(url=url, headers=headers)
        else:
            raise ConstellationInvalidMethodException(method, path)

        if response.status_code != HTTPStatus.OK:
            raise ConstellationException(
                f"Unexpected non-200 response from constellation endpoint {url}: {response.text}",
                response.status_code,
            )

        if field_to_retrieve:
            field_value = response.json().get(field_to_retrieve)
            if not field_value:
                raise ConstellationInvalidResponseException(
                    f"Could not retrieve field {field_to_retrieve} from request to {path}",
                    response.text,
                )
            return field_value

        return response.json()

    def run(self, func: Function) -> SLLResult:
        """Primary method for interacting with the SLL API. This method submits
        an experiment/analysis/simulation on AWS, and quickly returns if the
        request was successful. The data from the result is not immediately
        stored, and is instead lazily queried when you need it.

        :param func: The object representing the analysis/simulation/experiment function.
        :type func: Function
        :returns: A result object that contains the data from the function evaluation in the `result`, `error`,
        and `messages` fields.
        :rtype: SLLResult
        """
        # create the payload from the function object, making sure to jsonify function
        # this is b/c the function json string is to be passed through the go endpoint, and into MM kernels
        payload = {"function": json.dumps(asdict(func.payload()))}
        # construct the request
        url = self.host() + "/obj/sll/run_function"
        headers = {
            "Authorization": "Bearer " + self._auth_token,
            "Content-Type": "application/json",
            "X-ECL-NotebookId": self._notebook_id,
        }
        resp = requests.post(url=url, data=payload, headers=headers)
        resp = self._run_api_function(url=url, payload=payload, headers=headers)

        # handle non-200 responses in a heavy-handed way for now
        if resp.status_code != 200:
            try:
                message = resp.json().get("message")
            except Exception:
                raise ValueError("Something went wrong on the API backend.")
            raise ValueError(f"Error occurred while submitting command to API. Message: {message}")
        command = resp.json()["command"]
        return SLLResult(command=Object(command["id"]), client=self)

    def get_results(self, command: Object) -> ResultPayload:
        # run the get request
        url = self.host() + "/obj/sll/function_results?command=" + command.id
        headers = {
            "Authorization": "Bearer " + self._auth_token,
            "Content-Type": "application/json",
            "X-ECL-NotebookId": self._notebook_id,
        }
        resp = self._get_api_results(url=url, headers=headers)

        # handle errors
        self._handle_get_results_errors(resp)

        # safe to try to unpack results
        res = resp.json().get("result")

        # if result is None, then we'll have to do a wait-and-poll again flow
        total_sleep_time = 0
        # 2 ** counter generator: e.g. first val is 1, second is 2, third is 4, etc.
        backoff_generator = backoff.expo()
        while res is None:
            if total_sleep_time > self._max_sleep_time:
                raise ValueError(
                    f"Result not found after waiting for more than {self._max_sleep_time / 60.} minutes. "
                    "Try again later."
                )
            # apply jitter to prevent any concurrent race conditions
            sleep = self._exp_backoff_time(backoff_generator)
            time.sleep(sleep)
            resp = requests.get(url=url, headers=headers)
            res = resp.json().get("result")
            total_sleep_time += sleep
        return ResultPayload(**json.loads(res))

    def _exp_backoff_time(self, backoff_iter: Generator) -> float:
        # NOTE: this initialization needs to happen with backoff 2.x.x otherwise the
        # generator returns None on next calls
        backoff_iter.send(None)
        # apply jitter to prevent any concurrent race conditions
        return self._sleep_time * backoff.full_jitter(next(backoff_iter))

    def _handle_get_results_errors(self, resp: requests.Response):
        # handle non-200 responses in a heavy-handed way for now
        if resp.status_code != 200:
            try:
                message = resp.json().get("message")
            except Exception:
                raise ValueError("Something went wrong on the API backend.")
            raise ValueError(f"Error occurred while submitting command to API. Message: {message}")

    @_retry_requests_decorator
    def _run_api_function(self, url: str, payload: dict, headers: dict) -> requests.Response:
        return requests.post(url=url, data=json.dumps(payload), headers=headers)

    @_retry_requests_decorator
    def _get_api_results(self, url: str, headers: dict) -> requests.Response:
        return requests.get(url=url, headers=headers)


def needs_response(fn):
    @functools.wraps(fn)
    def wrapper(self: SLLResult, *args, **kwargs):
        if self._response is None:
            payload = self._client.get_results(self._command)
            self._response = SLLResult.Response(payload=payload, result=deserialize_item(payload.result))
        return fn(self, *args, **kwargs)

    return wrapper


class SLLResult:
    @dataclass
    class Response:
        payload: ResultPayload
        result: ListableType

    def __init__(self, command: Object, client: Constellation):
        self._command = command
        self._client = client

        self._response: SLLResult.Response | None = None

    def __repr__(self) -> str:
        default = "<Unevaluated>"
        result = self._response.result if self._response else default
        error = self._response.payload.error if self._response else default
        messages = self._response.payload.messages if self._response else default
        return f"SLLResult(result={result}, error={error}, messages={messages})"

    @property
    @needs_response
    def result(self) -> ListableType | None:
        assert self._response
        return self._response.result

    @property
    @needs_response
    def error(self) -> bool | None:
        assert self._response
        return self._response.payload.error

    @property
    @needs_response
    def messages(self) -> list[str] | None:
        assert self._response
        return self._response.payload.messages



import datetime
import json
import logging
import os
import random
import string
from zoneinfo import ZoneInfo

import boto3
import pytest

from pysll import Constellation


@pytest.fixture(scope="session")
def client():
    """Helper function to retrieve an actively logged in client.

    This uses AWS secrets manager to retrieve credentials.  If you do
    not have access to ECL'S AWS secrets manager, you must override this
    function in order to run the tests.
    """

    # Now, get an auth token by logging into constellation for each environment
    client = Constellation(host="https://constellation-stage.emeraldcloudlab.com")
    try:
        with open("./.auth") as handle:
            auth_token, notebook_id = handle.read().strip().split("\n")
    except FileNotFoundError:
        secrets_client = boto3.session.Session().client(service_name="secretsmanager")  # type: ignore
        credentials_secret = secrets_client.get_secret_value(SecretId="manifold-service-user-constellation-credentials")
        credentials_dict = json.loads(credentials_secret["SecretString"])
        username = credentials_dict["username"]
        password = credentials_dict["password"]
        client.login(username, password)
    else:
        logging.warning("Loaded credentials from .auth file")
        client._auth_token = auth_token
        client._notebook_id = notebook_id

    return client


@pytest.fixture(scope="session")
def sample_file():
    test_file_path = f"cloud-file-test_{''.join(random.choices(string.ascii_lowercase, k=12))}.txt"

    entropy = "".join(random.choices(string.printable, k=30))
    contents = "\n".join(
        [
            "cloud file test!",
            f"generated on: {datetime.datetime.now(tz=ZoneInfo('UTC')).isoformat()}",
            f"here's some entropy: {entropy}",
        ]
    )
    with open(test_file_path, "w") as f:
        f.write(contents)

    yield test_file_path, contents

    try:
        os.remove(test_file_path)
    except FileNotFoundError:
        pass



import logging
import pathlib

from pysll import Constellation
from pysll.models import Object, VariableUnitValue


def test_simple_upload(client: Constellation):
    data = client.upload("Object.Example.Data", None, {})

    assert data["new_object"]
    assert data["type"] == "Object.Example.Data"


def test_upload_public_object(client: Constellation):
    data = client.upload("Object.Example.Data", None, {}, allow_public_objects=True)

    object_id = data["id"]
    notebook = client.download(Object(object_id), "Notebook")

    assert notebook is None


def test_single_upload_cloud_file(client, sample_file):
    """Create a dummy file and upload it as a cloud file."""

    file_name, contents = sample_file
    assert contents

    logging.info(f"uploading file {file_name} to cloud...")
    logging.info(contents)

    response = client.upload_cloud_file(file_name, None)
    resolved_object = response["resolved_object"]

    download_response = client.download(
        Object(resolved_object["id"]), ["id", "FileSize", "FileName", "FileType", "type"]
    )
    assert download_response[0] == resolved_object["id"]
    assert download_response[1] == VariableUnitValue(len(contents), "Bytes")
    assert download_response[2] == pathlib.Path(file_name).stem
    assert download_response[3] == '"txt"'
    assert download_response[4] == "Object.EmeraldCloudFile"
    assert "Notebook" not in download_response


def test_single_upload_cloud_file_with_notebook(client: Constellation, sample_file: tuple[str, str]):
    """Create a dummy file and upload it as a cloud file."""

    file_name, contents = sample_file
    assert contents

    logging.info(f"uploading file {file_name} to cloud...")
    logging.info(contents)

    response = client.upload_cloud_file(file_name, None)
    resolved_object = response["resolved_object"]

    download_response = client.download(
        Object(resolved_object["id"]), ["id", "FileSize", "FileName", "FileType", "type"]
    )
    assert download_response[0] == resolved_object["id"]
    assert download_response[1] == VariableUnitValue(len(contents), "Bytes")
    assert download_response[2] == pathlib.Path(file_name).stem
    assert download_response[3] == '"txt"'
    assert download_response[4] == "Object.EmeraldCloudFile"
    assert "Notebook" not in download_response



import datetime
import os
from operator import attrgetter

import pytest

from pysll.exceptions import ConstellationObjectDoesNotExistException
from pysll.models import Model, Object, VariableUnitValue


def test_simple_download(client):
    """Test a simple download on a single object."""
    assert client.download(Object("id:Z1lqpMzvkGMV"), "Name") == "bsmith"


def test_multiple_objects_download(client):
    """Test a simple download on multiple objects."""
    assert client.download([Object("id:Z1lqpMzvkGMV"), Object("id:o1k9jAGmo6Jm")], "Name") == [
        "bsmith",
        "pavan.shah",
    ]


def test_multiple_field_download(client):
    """Test a simple download on multiple fields."""
    assert client.download(Object("id:Z1lqpMzvkGMV"), ["Name", "FirstName"]) == [
        "bsmith",
        "Ben",
    ]


def test_multiple_field_multiple_object_download(client):
    """Test a simple download multiple objects and fields."""
    assert client.download([Object("id:Z1lqpMzvkGMV"), Object("id:o1k9jAGmo6Jm")], ["Name", "FirstName"]) == [
        ["bsmith", "Ben"],
        ["pavan.shah", "Pavan"],
    ]


def test_download_non_existant_object(client):
    """Test a simple download on a non-existant object."""
    with pytest.raises(ConstellationObjectDoesNotExistException):
        client.download(Object("id:abc123"), "Name")


def test_download_non_existant_field(client):
    """Test a simple download on a non-existant field."""
    assert client.download(Object("id:Z1lqpMzvkGMV"), "FakeField") is None


def test_download_empty_field(client):
    """Test a simple download on an empty field."""
    assert client.download(Object("id:Z1lqpMzvkGMV"), "Biography") is None


def test_download_empty_link_field(client):
    """Test a simple download on an empty link field."""
    assert client.download(Object("id:Z1lqpMzvkGMV"), "ProtocolStack") is None


def test_download_multiple_link_field(client):
    """Test a simple download on a non-existant field."""
    assert client.download(Object("id:Z1lqpMzvkGMV"), "FinancingTeams") == [Object("id:1ZA60vLeXa7a")]


def test_download_all(client):
    """Test downloading all fields on an object."""
    all_dict = client.download(Object("id:Z1lqpMzvkGMV"))
    assert all_dict["Name"] == "bsmith"


def test_download_parts(client):
    """Test downloading parts of a field."""
    assert client.download(Object("id:Z1lqpMzvkGMV"), "ProtocolsAuthored[[1]]") == Object("id:wqW9BP7dkv8G")


def test_download_parts_traversal(client):
    """Test downloading parts of a field then traversing a link."""
    assert client.download(Object("id:Z1lqpMzvkGMV"), "ProtocolsAuthored[[1]][Author]") == Object("id:Z1lqpMzvkGMV")


def test_download_parts_column(client):
    """Test downloading columns of a field."""
    result = client.download(Object("id:o1k9jAKpnRpm"), "Contents[[All,1]]")
    assert isinstance(result, list)
    assert all(isinstance(item, str) for item in result)


def test_download_parts_rows_column(client):
    """Test downloading rows and columns of a field."""
    assert client.download(Object("id:o1k9jAKpnRpm"), "Contents[[1,1]]") == "A1"


def test_download_parts_column_traversal(client):
    """Test downloading via traversal and then fetching columns of a field."""
    result = client.download(Object("id:01G6nvDGrlWE"), "ContainersOut[Contents][[All,1]]")
    assert isinstance(result, list)
    assert len(result) == 4
    assert all(all(isinstance(item, str) for item in subarray) for subarray in result)


def test_download_parts_row_column_traversal(client):
    """Test downloading via traversal and then fetching columns of a field."""
    assert client.download(Object("id:01G6nvDGrlWE"), "ContainersOut[Contents][[1,1]]") == ["A1", "A1", "A1", "A1"]


def test_download_parts_specific_row_column_traversal(client):
    """Test downloading via traversal and then fetching columns of a field."""
    result = client.download(Object("id:01G6nvDGrlWE"), "ContainersOut[[1]][Contents][[All,1]]")
    assert isinstance(result, list)
    assert all(isinstance(item, str) for item in result)


# Downloading Recursively


def test_download_repeated_single_field(client):
    """Test that you can download recursively from a single field."""
    assert client.download(Object("id:4pO6dM57V6K7"), "Repeated[Container]") == [Object("id:xRO9n3BdWrxY")]


def test_download_repeated_single_field_with_traversal(client):
    """Test that you can download recursively from a single field with a
    traversal."""
    assert client.download(Object("id:4pO6dM57V6K7"), "Repeated[Container][Model]") == [Model("id:J8AY5jDwb4B9")]


def test_download_repeated_multiple_field(client):
    """Test that you can download recursively from a multiple field."""
    result = client.download(Object("id:KBL5DvPoNP0a"), "Repeated[Subprotocols]")
    assert isinstance(result, list)
    assert [arr[0] for arr in result[:3]] == [
        Object("id:GmzlKjNBxv7E"),
        Object("id:n0k9mGOWAmXo"),
        Object("id:o1k9jAoWJLk4"),
    ]


def test_download_repeated_multiple_field_with_traversal(client):
    """Test that you can download recursively from a multiple field with a
    traversal."""
    result = client.download(Object("id:KBL5DvPoNP0a"), "Repeated[Subprotocols][DateCreated]")
    assert isinstance(result, list)
    assert [isinstance(item[0], datetime.datetime) for item in result]


def test_download_repeated_buried_traversal(client):
    """Test that you can download recursively from a buried Repeated
    reference."""
    result = client.download(Object("id:4pO6dM5GvK0X"), "Resources[Repeated[Container]][Model]")
    assert isinstance(result, list)
    assert all(len(item) == 1 for item in result)
    assert [isinstance(item[0], Object) for item in result]


def test_download_repeated_multiple_field_with_column(client):
    """Test that you can download recursively from a multiple field that
    required a specific column."""
    result = client.download(Object("id:Z1lqpMz63avW"), "Repeated[Contents[[All,2]]]")
    assert result == [
        [Object("id:o1k9jAGJzXDm"), None],
        [Object("id:GmzlKjPkdpVe"), None],
        [Object("id:8qZ1VW0Mb5OZ"), [[Object("id:M8n3rx0lqXz8"), None], [Object("id:Z1lqpMz63PAo"), None]]],
    ]


def test_download_repeated_multiple_field_with_column_and_traversal(client):
    """Test that you can download recursively from a multiple field that
    required a specific column and then a field traversal."""
    result = client.download(Object("id:Z1lqpMz63avW"), "Repeated[Contents[[All,2]]][Model]")
    assert result == [
        [Model("id:wqW9BP7RBNZG"), None],
        [Model("id:8qZ1VWNw1z0X"), None],
        [Model("id:E8zoYveRll17"), [[Model("id:8qZ1VWNmdLBD"), None], [Model("id:8qZ1VWNmdLBD"), None]]],
    ]


def test_download_repeated_listed_objects(client):
    """Test that you can download from a list of objects with a repeated field
    reference."""
    result = client.download([Object("id:R8e1PjRDvAAK"), Object("id:AEqRl9540rjd")], "Repeated[Subprotocols]")
    assert result[0] == [
        [Object("id:bq9LA0dBD86a"), None],
        [Object("id:n0k9mGzReVYr"), [[Object("id:BYDOjv1VZ0Zr"), None]]],
    ]
    assert result[1] == [
        [Object("id:kEJ9mqaVW6NX"), None],
        [Object("id:jLq9jXY4Bdx1"), [[Object("id:dORYzZn08jnD"), None]]],
    ]


def test_download_repeated_listed_objects_fields(client):
    """Test that you can download from a list of objects with a list of fields
    including a repeated field reference."""
    result = client.download([Object("id:R8e1PjRDvAAK"), Object("id:AEqRl9540rjd")], ["Name", "Repeated[Subprotocols]"])
    assert result[0] == [
        "LegacyID:42",
        [
            [Object("id:bq9LA0dBD86a"), None],
            [Object("id:n0k9mGzReVYr"), [[Object("id:BYDOjv1VZ0Zr"), None]]],
        ],
    ]
    assert result[1] == [
        "LegacyID:43",
        [
            [Object("id:kEJ9mqaVW6NX"), None],
            [Object("id:jLq9jXY4Bdx1"), [[Object("id:dORYzZn08jnD"), None]]],
        ],
    ]


# Downloading links


def test_chained_download(client):
    """Test that you can chain downloads off of links."""
    assert client.download(client.download(Object("id:Z1lqpMzvkGMV"), "FinancingTeams"), "Name") == ["Engineering"]


def test_single_traversal(client):
    """Test that you can traverse links."""
    assert client.download(Object("id:Z1lqpMzvkGMV"), "FinancingTeams[Name]") == ["Engineering"]


def test_multiple_traversal(client):
    """Test that you can perform multiple link traversals in a single
    request."""
    assert client.download(Object("id:Z1lqpMzvkGMV"), "FinancingTeams[BillingAddress][Name]") == ["ECL-1"]


def test_multiple_multiple_field(client):
    """Test that you can download a multiple field."""
    field_value = client.download(Object("id:Z1lqpMzvkGMV"), "SafetyTrainingLog")
    assert len(field_value) == 3
    assert field_value[0][1] == "ChemicalHandling"
    assert isinstance(field_value[0][0], datetime.datetime)


def test_multiple_field_single_field_traversal(client):
    """Test that you can download a single field through a multiple field."""
    field_value = client.download(Object("id:Z1lqpMzvkGMV"), "DirectReports[Name]")
    assert field_value == ["tommy.harrelson", "ayi", "brad"]


def test_multiple_field_multiple_field_traversal(client):
    """Test that you can download a multiple field through a multiple field."""
    field_value = client.download(Object("id:Z1lqpMzvkGMV"), "DirectReports[SafetyTrainingLog]")
    assert len(field_value) == 3
    assert field_value[0] is None
    assert len(field_value[1]) == 3
    assert len(field_value[2]) == 3
    assert len(field_value[1][0]) == 3
    assert field_value[1][0][1] == "ChemicalHandling"
    assert isinstance(field_value[1][1][0], datetime.datetime)


# Downloading various types


def test_download_integer(client):
    """Test that integer field types are correctly formatted."""
    assert client.download(Object("id:BYDOjvG4l3Ol"), "InjectionIndex") == 28


def test_download_string(client):
    """Test that string field types are correctly formatted."""
    assert client.download(Object("id:BYDOjvG4l3Ol"), "DataType") == "Analyte"


def test_download_quantity(client):
    """Test that quantity field types are correctly formatted."""
    assert client.download(Object("id:BYDOjvG4l3Ol"), "SampleVolume") == VariableUnitValue(10, "Microliters")


def test_download_date(client):
    """Test that date fields are correctly formatted."""
    # Note - retrieving the default units automatically is not yet supported
    assert client.download(Object("id:BYDOjvG4l3Ol"), "DateInjected") == datetime.datetime(2022, 1, 7, 17, 12, 19)


def test_file_download_with_automatic_download(client):
    """Test that files are automatically downloaded when requested."""
    cloud_file = client.download(Object("id:Z1lqpMzvkGMV"), "PhotoFile", auto_download_cloud_files=True)
    assert os.path.exists(cloud_file.local_path)


def test_file_download_with_automatic_download_and_byte_size(client):
    """Test that files are automatically downloaded when requested."""
    cloud_file = client.download(
        Object("id:Z1lqpMzvkGMV"), "PhotoFile", auto_download_cloud_files=True, byte_size=10_000_000
    )
    assert os.path.exists(cloud_file.local_path)


def test_file_download_without_automatic_download(client):
    """Test that files are not automatically downloaded when not requested."""
    cloud_file = client.download(Object("id:Z1lqpMzvkGMV"), "PhotoFile")
    assert cloud_file == Object("id:KBL5Dvw60ve7")
    assert cloud_file.local_path == ""


def test_blob_ref_download(client):
    """Test that blob refs are downloaded automatically correctly."""
    absorbance_data = client.download(Object("id:BYDOjvG4l3Ol"), "Absorbance")
    assert len(absorbance_data) == 7201
    assert len(absorbance_data[3]) == 2
    assert absorbance_data[3][0].unit == "'Minutes'"
    assert absorbance_data[3][1].unit == "'Milli' 'AbsorbanceUnit'"


def test_absorbance_download(client):
    """Test that Absorbance fields are downloaded automatically correctly.

    Absorbance fields are given by BigQuantityArrays (e.g. blobrefs
    under the hood).
    """
    absorbance_data = client.download(Object("id:9RdZXvdLrJwa"), "Absorbance")
    assert len(absorbance_data) == 2250
    assert len(absorbance_data[3]) == 2
    assert absorbance_data[3][0].unit == "'Minutes'"
    assert absorbance_data[3][1].unit == "'Milli' 'AbsorbanceUnit'"


def test_multi_blob_ref_download(client):
    """Test that multiple blob refs are downloaded automatically correctly."""
    absorbance_data = client.download([Object("id:BYDOjvG4l3Ol"), Object("id:BYDOjvG4l3Ol")], "Absorbance")
    assert len(absorbance_data) == 2
    absorbance_data = absorbance_data[0]
    assert len(absorbance_data) == 7201
    assert len(absorbance_data[3]) == 2
    assert absorbance_data[3][0].unit == "'Minutes'"
    assert absorbance_data[3][1].unit == "'Milli' 'AbsorbanceUnit'"


def test_blob_ref_download_with_byte_size(client):
    """Test that blob refs are downloaded automatically correctly."""
    absorbance_data = client.download(Object("id:BYDOjvG4l3Ol"), "Absorbance", byte_size=10_000_000)
    assert len(absorbance_data) == 7201
    assert len(absorbance_data[3]) == 2
    assert absorbance_data[3][0].unit == "'Minutes'"
    assert absorbance_data[3][1].unit == "'Milli' 'AbsorbanceUnit'"


def test_quantity_array_download(client):
    """Test that quantity arrays are downloaded correctly."""
    scatter_info = client.download(Object("id:BYDOjvG4l3Ol"), "Scattering")
    assert isinstance(scatter_info, list)
    assert len(scatter_info) == 361


def test_variable_unit_download(client):
    """Test that variable units are downloaded correctly."""
    composition = client.download(Object("id:O81aEB16GlJ1"), "Composition")
    assert isinstance(composition, list)
    assert len(composition) == 2
    assert isinstance(composition[0], list)
    assert len(composition[0]) == 3
    assert isinstance(composition[0][0], VariableUnitValue)
    assert composition[0][1] == Model("id:E8zoYvN6m61A")


def test_association_download(client):
    """Test that associations are downloaded correctly."""
    composition = client.download(Object("id:XnlV5jKZwmp3"), "ResolvedOptions")
    assert isinstance(composition, dict)
    assert len(composition) == 295
    assert isinstance(composition["AbsorbanceSamplingRate"], VariableUnitValue)
    assert composition["AbsorbanceSamplingRate"].value == 20
    assert composition["Instrument"] == Object("id:wqW9BP4ARZVw")


def test_download_cloud_file(client):
    """Test that cloud file download correctly.

    This is the PhotoFile of Object("id:Z1lqpMzvkGMV").
    """
    cloud_file_path = client.download_cloud_file(Object("id:KBL5Dvw60ve7"))
    assert os.path.exists(cloud_file_path)


def test_download_cloud_file_with_byte_size(client):
    """Test that cloud file download correctly.

    This is the PhotoFile of Object("id:Z1lqpMzvkGMV").
    """
    cloud_file_path = client.download_cloud_file(Object("id:KBL5Dvw60ve7"), byte_size=10_000_000)
    assert os.path.exists(cloud_file_path)


def test_download_model(client):
    assert client.download(Object("id:qdkmxzGkAK0a"), "Container[Model]") == Model(
        id="id:L8kPEjkmLbvW", type="Model.Container.Plate"
    )


def test_resolve_type_mutates_objects(client):
    sample = Object("id:qdkmxzGkAK0a")
    _ = client.resolve_type(sample)
    assert sample.type == "Object.Sample"


def test_resolve_type_for_multiple_objects(client):
    assert [
        obj.type
        for obj in [
            client.resolve_type(kind)
            for kind in [
                Object("id:Z1lqpMzvkGMV"),
                Object("id:o1k9jAGmo6Jm"),
                Model("id:L8kPEjkmLbvW"),
            ]
        ]
    ] == [
        "Object.User.Emerald.Developer",
        "Object.User.Emerald.Developer",
        "Model.Container.Plate",
    ]


def test_resolve_type_overwrites_existing_type(client):
    obj = Object("id:Z1lqpMzvkGMV", type="Some.Crazy.Type")
    assert obj.type == "Some.Crazy.Type"
    _ = client.resolve_type(obj)
    assert obj.type == "Object.User.Emerald.Developer"


def test_resolve_type_listable_overload(client):
    objects = client.resolve_type(
        [
            Object("id:Z1lqpMzvkGMV"),
            Object("id:o1k9jAGmo6Jm"),
            Model("id:L8kPEjkmLbvW"),
        ]
    )
    [a, b, c] = objects
    assert (a.type, b.type, c.type) == (
        "Object.User.Emerald.Developer",
        "Object.User.Emerald.Developer",
        "Model.Container.Plate",
    )

    source = [Object("id:Z1lqpMzvkGMV"), Object("id:o1k9jAGmo6Jm"), Model("id:L8kPEjkmLbvW")]
    _ = client.resolve_type(iter(source))
    with pytest.raises(StopIteration):
        next(_)

    assert list(map(attrgetter("type"), source)) == [
        "Object.User.Emerald.Developer",
        "Object.User.Emerald.Developer",
        "Model.Container.Plate",
    ]


def test_resolve_type_on_non_existing_object(client):
    with pytest.raises(ConstellationObjectDoesNotExistException):
        _ = client.resolve_type(
            [
                Object("id:Z1lqpMzvkGMV"),
                Object("id:o1k9jAGmo6Jm"),
                Object("id:the_intruder"),
                Model("id:L8kPEjkmLbvW"),
            ]
        )



import backoff

from pysll.constellation import Constellation, SLLResult
from pysll.functions import AnalyzePeaks
from pysll.models import (
    Expression,
    Function,
    Object,
    Option,
    Quantity,
    ResultPayload,
    Symbol,
    deserialize_item,
    serialize_item,
)
from pysll.unit_operations import Transfer


def test_api_function_construction():
    """Test that checks we can properly build the Function class from a variety
    of sources, and we can serialize it into dicts."""
    func = Function([[1, 2], [2, 5], [3, 1]])
    assert func.inputs == [[[1, 2], [2, 5], [3, 1]]]
    assert func.options == []

    # test out building with an option
    func = Function([[1, 2], [2, 5], [3, 1]], AbsoluteThreshold=10.4)
    assert func.inputs == [[[1, 2], [2, 5], [3, 1]]]
    assert func.options == [Option(name="AbsoluteThreshold", value=10.4)]

    # make an api dict to check if we can build a function from that
    func = Function.from_dict(
        {
            "name": "AnalyzePeaks",
            "data_type": "function",
            "inputs": [
                [
                    [{"data_type": "float", "value": 1}, {"data_type": "float", "value": 2}],
                    [{"data_type": "float", "value": 2}, {"data_type": "float", "value": 5}],
                    [{"data_type": "float", "value": 3}, {"data_type": "float", "value": 1}],
                ]
            ],
            "options": [
                {"data_type": "option", "name": "AbsoluteThreshold", "value": {"data_type": "float", "value": 10.4}}
            ],
        }
    )
    assert type(func).__name__ == "AnalyzePeaks"
    assert func.inputs == [[[1, 2], [2, 5], [3, 1]]]
    assert func.options == [Option(name="AbsoluteThreshold", value=10.4)]


def test_unit_op_construction():
    """Test that checks we can properly build the unit operations from a
    variety of sources, and we can serialize it into dicts."""

    # make an api dict to check if we can build a function from that
    func = Function.from_dict(
        {
            "name": "Transfer",
            "data_type": "function",
            "inputs": [],
            "options": [
                {"data_type": "option", "name": "AbsoluteThreshold", "value": {"data_type": "float", "value": 10.4}}
            ],
        }
    )
    assert type(func).__name__ == "Transfer"
    assert func.options == [Option(name="AbsoluteThreshold", value=10.4)]


def test_api_function_deserialization():
    """Test that checks we can properly deserialize a Function class into a
    dict."""
    # test out building with an option
    func = Function([[1, 2], [2, 5], [3, 1]], AbsoluteThreshold=10.4)
    func_dict = serialize_item(func)
    assert isinstance(func_dict, dict)
    assert func_dict["name"] == "Function"
    assert func_dict["inputs"] == [
        [
            [{"data_type": "integer", "value": 1}, {"data_type": "integer", "value": 2}],
            [{"data_type": "integer", "value": 2}, {"data_type": "integer", "value": 5}],
            [{"data_type": "integer", "value": 3}, {"data_type": "integer", "value": 1}],
        ]
    ]
    assert func_dict["options"] == [
        {"data_type": "option", "name": "AbsoluteThreshold", "value": {"data_type": "float", "value": 10.4}}
    ]
    assert func_dict["data_type"] == "function"

    # test it out with an object input
    func = Function(Object("id:xyz", type="Object.Data.Chromatography"), OptionName="blah")
    func_dict = serialize_item(func)
    assert isinstance(func_dict, dict)
    assert len(func_dict["inputs"]) == 1
    inp = list(func_dict["inputs"])[0]
    assert isinstance(inp, dict)
    assert inp["data_type"] == "object"
    assert inp.get("id") == "id:xyz"
    assert inp.get("name") is None
    assert inp.get("type") == "Object.Data.Chromatography"

    # check the options
    assert len(func_dict["options"]) == 1
    op = list(func_dict["options"])[0]
    assert isinstance(op, dict)
    assert op["data_type"] == "option"
    assert op["name"] == "OptionName"
    assert op["value"] == {"data_type": "string", "value": "blah"}


def test_sll_result(client):
    """Test that an SLL object can be created without failure."""
    res = SLLResult(command=Object("id:abc"), client=client)
    assert res._command.id == "id:abc"
    assert isinstance(res._client, Constellation)


def test_analyze_peaks_from_python():
    """Test that AnalyzePeaks works as expected."""
    peaks = AnalyzePeaks(Object("id:abc"))
    assert isinstance(peaks.inputs, list)
    assert len(peaks.inputs) == 1
    assert isinstance(peaks.inputs[0], Object)
    assert peaks.inputs[0].id == "id:abc"

    # try it out with options
    peaks = AnalyzePeaks(Object("id:abc"), AbsoluteThreshold=10.34, HighlightBaselinePoints=True)
    assert isinstance(peaks.options, list)
    assert len(peaks.options) == 2
    assert isinstance(peaks.options[0], Option)
    assert peaks.options[0].value == 10.34
    assert isinstance(peaks.options[1], Option)
    assert peaks.options[1].value is True


def test_unit_ops_from_python():
    """Test that AnalyzePeaks works as expected."""
    t = Transfer(Source=Object("id:xyz"))
    assert isinstance(t.inputs, list)
    assert len(t.inputs) == 0
    assert isinstance(t.options, list)
    assert len(t.options) == 1
    assert isinstance(t.options[0], Option)
    assert isinstance(t.options[0].value, Object)
    assert t.options[0].value.id == "id:xyz"


def test_analyze_peaks_from_dict():
    """Test that creating AnalyzePeaks from dict works as expected."""
    peaks = Function.from_dict(
        {
            "data_type": "function",
            "name": "AnalyzePeaks",
            "inputs": [{"data_type": "object", "id": "id:xyz", "type": "Object.Data.Chromatography", "name": None}],
            "options": [
                {"data_type": "option", "name": "AbsoluteThreshold", "value": {"data_type": "float", "value": 10.34}},
                {
                    "data_type": "option",
                    "name": "HighlightBaselinePoints",
                    "value": {"data_type": "bool", "value": True},
                },
            ],
        }
    )
    assert isinstance(peaks, AnalyzePeaks)
    assert isinstance(peaks.inputs, list)
    assert len(peaks.inputs) == 1
    assert isinstance(peaks.inputs[0], Object)
    assert peaks.inputs[0].id == "id:xyz"

    # test the options
    assert isinstance(peaks.options, list)
    assert len(peaks.options) == 2
    assert isinstance(peaks.options[0], Option)
    assert peaks.options[0].value == 10.34
    assert isinstance(peaks.options[1], Option)
    assert peaks.options[1].value is True


def test_failed_dtype_analyze_peaks_from_dict():
    """Test that creating AnalyzePeaks from dict fails as expected when given a
    bad input."""
    try:
        res = Function.from_dict(
            {
                # make a mistake in the top level datatype
                "data_type": "func",
                "name": "AnalyzePeaks",
                "inputs": [{"data_type": "object", "id": "id:xyz", "type": "Object.Data.Chromatography", "name": None}],
                "options": [
                    {
                        "data_type": "option",
                        "name": "AbsoluteThreshold",
                        "value": {"data_type": "float", "value": 10.34},
                    },
                    {
                        "data_type": "option",
                        "name": "HighlightBaselinePoints",
                        "value": {"data_type": "bool", "value": True},
                    },
                ],
            }
        )
    except TypeError as e:
        res = e
    assert isinstance(res, TypeError)


def test_failed_obj_dtype_analyze_peaks_from_dict():
    try:
        res = Function.from_dict(
            {
                "data_type": "function",
                "name": "AnalyzePeaks",
                # make a mistake in the object level datatype
                "inputs": [{"data_type": "obj", "id": "id:xyz", "type": "Object.Data.Chromatography", "name": None}],
                "options": [
                    {
                        "data_type": "option",
                        "name": "AbsoluteThreshold",
                        "value": {"data_type": "float", "value": 10.34},
                    },
                    {
                        "data_type": "option",
                        "name": "HighlightBaselinePoints",
                        "value": {"data_type": "bool", "value": True},
                    },
                ],
            }
        )
    except TypeError as e:
        res = e
    assert isinstance(res, TypeError)


def test_failed_option_val_analyze_peaks_from_dict():
    try:
        res = Function.from_dict(
            {
                "data_type": "function",
                "name": "AnalyzePeaks",
                "inputs": [{"data_type": "object", "id": "id:xyz", "type": "Object.Data.Chromatography", "name": None}],
                "options": [
                    {
                        "data_type": "option",
                        "name": "AbsoluteThreshold",
                        # make a mistake in the value of the option
                        "value": {
                            "data_type": "float",
                            "value": {"data_type": "object", "type": "Object.Data", "id": "id:abc", "name": None},
                        },
                    },
                    {
                        "data_type": "option",
                        "name": "HighlightBaselinePoints",
                        "value": {"data_type": "bool", "value": True},
                    },
                ],
            }
        )
    except TypeError as e:
        res = e
    assert isinstance(res, TypeError)


# test out all the serializations and deserializations except functions, which are tested above
def test_serialize_float():
    assert serialize_item(10.1) == {"data_type": "float", "value": 10.1}


def test_serialize_int():
    assert serialize_item(10) == {"data_type": "integer", "value": 10}


def test_serialize_bool():
    assert serialize_item(True) == {"data_type": "bool", "value": True}


def test_serialize_string():
    assert serialize_item("$Failed") == {"data_type": "string", "value": "$Failed"}


def test_serialize_object():
    assert serialize_item(Object("id:xyz")) == {"data_type": "object", "id": "id:xyz", "type": "", "name": None}


def test_serialize_option():
    assert serialize_item(Option(name="blah", value=10)) == {
        "data_type": "option",
        "name": "blah",
        "value": {"data_type": "integer", "value": 10},
    }


def test_serialize_quantity():
    assert serialize_item(Quantity(value=10, units="Meters")) == {
        "data_type": "quantity",
        "value": 10,
        "units": "Meters",
    }


def test_serialize_expression():
    assert serialize_item(Expression(string="dog")) == {"data_type": "expression", "string": "dog"}


def test_serialize_symbol():
    assert serialize_item(Symbol(name="dog")) == {"data_type": "expression", "string": "dog"}


def test_deserialize_float():
    assert deserialize_item({"data_type": "float", "value": 10.1}) == 10.1


def test_deserialize_int():
    assert deserialize_item({"data_type": "integer", "value": 10}) == 10


def test_deserialize_bool():
    assert deserialize_item({"data_type": "bool", "value": True}) is True


def test_deserialize_string():
    assert deserialize_item({"data_type": "string", "value": "$Failed"}) == "$Failed"


def test_deserialize_object():
    assert deserialize_item({"data_type": "object", "id": "id:xyz", "type": "", "name": None}) == Object(id="id:xyz")


def test_deserialize_option():
    assert deserialize_item(
        {
            "data_type": "option",
            "name": "blah",
            "value": {"data_type": "integer", "value": 10},
        }
    ) == Option(name="blah", value=10)


def test_deserialize_quantity():
    assert deserialize_item(
        {
            "data_type": "quantity",
            "value": 10,
            "units": "Meters",
        }
    ) == Quantity(value=10, units="Meters")


def test_deserialize_expression():
    assert deserialize_item({"data_type": "expression", "string": "dog"}) == Expression(string="dog")


def test_sll_result_repr(client, mocker):
    res = SLLResult(Object("id:xyz"), client)
    assert str(res) == "SLLResult(result=<Unevaluated>, error=<Unevaluated>, messages=<Unevaluated>)"

    mocker.patch(
        "pysll.constellation.Constellation.get_results",
        return_value=ResultPayload(
            result={"data_type": "object", "type": "Object.Analysis.Peaks", "name": None, "id": "id:xyz"},
            error=False,
            messages=[],
        ),
    )

    assert res.error is False  # this will trigger the call to get results
    assert res.messages == []
    assert str(res) == 'SLLResult(result=Object[Analysis, Peaks, "id:xyz"], error=False, messages=[])'


def test_sll_result_failed_repr(client, mocker):
    res = SLLResult(Object("id:xyz"), client)
    assert str(res) == "SLLResult(result=<Unevaluated>, error=<Unevaluated>, messages=<Unevaluated>)"

    mocker.patch(
        "pysll.constellation.Constellation.get_results",
        return_value=ResultPayload(
            result={"data_type": "string", "value": "$Failed"},
            error=True,
            messages=["AnalyzePeaks::OptionError : FlowRate is not an option for AnalyzePeaks."],
        ),
    )

    assert res.error  # this will trigger the call to get results
    assert str(res) == (
        "SLLResult(result=$Failed, error=True, messages=['AnalyzePeaks::OptionError : "
        "FlowRate is not an option for AnalyzePeaks.'])"
    )


def test_sll_result_retry(client):
    backoff_gen = backoff.expo()
    t = client._exp_backoff_time(backoff_gen)
    assert isinstance(t, float)
    assert t > 0

    # try it again to make sure we can run subsequent backoff
    # time calls on the same generator
    t = client._exp_backoff_time(backoff_gen)
    assert isinstance(t, float)
    assert t > 0



import datetime
import hashlib
import threading
from functools import reduce
from typing import Any


def constellation_date_string(dt):
    """Return a datetime formatted in the format constellation
    wants as a string like: 2021-02-17T15:16:36Z

    Note - the datetime object must be in utc time."""
    return dt.strftime("%Y-%m-%dT%H:%M:%SZ")


def constellation_date_from_string(date_string):
    """Return a datetime object from the string format that constellation
    uses like: 2021-02-09T04:12:02.618767Z

    Note - the datetime object will be in utc time."""
    # Handle both second and subsecond precision on datetimes
    try:
        return datetime.datetime.strptime(date_string, "%Y-%m-%dT%H:%M:%S.%fZ")
    except ValueError:
        pass

    return datetime.datetime.strptime(date_string, "%Y-%m-%dT%H:%M:%SZ")


def parts_to_xml(parts):
    part_nodes = "\n".join(
        f"""
        <Part>
            <ETag>{part["ETag"]}</ETag>
            <PartNumber>{part["PartNumber"]}</PartNumber>
        </Part>
        """
        for part in parts
    )

    xml = f"""
        <?xml version="1.0" encoding="UTF-8"?>
        <CompleteMultipartUpload xmlns="http://s3.amazonaws.com/doc/2006-03-01/">
        {part_nodes}
        "</CompleteMultipartUpload>"
    """

    return xml.encode("ascii")


def md5_hash_file(file_path):
    with open(file_path, "rb") as file:
        chunk_size = 1_000_000  # 1MB, a bit arbitrary
        md5 = hashlib.md5()
        chunk = file.read(chunk_size)
        while chunk:
            md5.update(chunk)
            chunk = file.read(chunk_size)
        return md5


def create_one_way_link(linked_type: str, linked_id: str):
    """Given the type and id in an object, create a two-way link to that
    object."""
    return {
        "$Type": "__JsonLink__",
        "object": {"id": linked_id, "type": linked_type},
    }


def create_two_way_link(linked_type: str, linked_id: str, linked_field: str):
    """Given the type, id and field in an object, create a two-way link to that
    object."""
    return {
        "$Type": "__JsonLink__",
        "object": {"id": linked_id, "type": linked_type},
        "field": {"name": linked_field},
    }


def extend(*dicts: dict) -> dict:
    """Returns a dictionary that combines all dictionaries passed as arguments
    without mutating any of them.

    >>> extend({"foo": "bar"}, {"food": "taco"})
    {"foo": "bar", "food": "taco"}
    >>> extend({"food": "cake"}, {"food": "taco", "answer": 42})
    {"food": "taco", "answer": 42}
    """

    def fold(acc: dict, curr: dict) -> dict:
        acc.update(curr)
        return acc

    return reduce(fold, dicts, {})


def truthy(v: str | None) -> bool:
    """
    >>> set(map(truthy, ["1", "true", "True", "TRUE", "tRuE", "on", "ON"]))
    {True}
    >>> set(map(truthy, [None, "", " ", "0", "0.0", "false", "False", "FALSE", "fAlSe", "off", "OFF"]))
    {False}
    """
    v = (v or "").strip()
    if v.lower() in {"true", "on"}:
        return True
    if v.lower() in {"false", "off"}:
        return False
    try:
        return bool(float(v))
    except ValueError:
        return bool(v)


def tmap(fn, *collections, **kw_collections) -> list[Any]:
    """Concurrent map.

    Map each item in each of the collections/keyword collections with
    the provided function in a separate thread.
    `tmap` can drastically improve performance when compared to
    `map` but should only be used for IO-bound functions that have
    no side effects.
    >>> tmap(download, urls)
    [<url1_data>, <url2_data>, <url3_data>]
    >>> datasets = tmap(read_csv, files)

    To thread a function over multiple inputs, you can provide multiple
    collections as positional arguments.
    >>> tmap(download, urls, fields)
    [<url1_field1_data>, <url2_field2_data>, ...]
    Note that each of the above functional evaluations are of the form
    `download(urls[i], fields[i])` where `i` is the index. Note that
    each collection must be the same length.

    If e.g. `download` has keyword arguments, then you may input other
    collections as keyword arguments.
    >>> tmap(download, urls, field=fields)
    [<url1_field1_data>, <url2_field2_data>, ...]
    which evaluates as `download(urls[i], field=fields[i])` for each
    element in the collections provided.
    """

    # check that each of the collections are the same size
    # set n to be either the first collection length or the first kwarg collection length
    if len(collections) > 0:
        n = len(collections[0])
    else:
        n = len(tuple(kw_collections.values())[0])
    for c in collections:
        if len(c) != n:
            raise ValueError("The collections passed to tmap must have the same length.")
    for key in kw_collections.keys():
        if len(kw_collections[key]) != n:
            raise ValueError("The collections passed to tmap must have the same length.")
    payload = [None] * n

    def process(i):
        args = (c[i] for c in collections)
        kwargs = {key: c[i] for key, c in kw_collections.items()}
        payload[i] = fn(*args, **kwargs)

    threads = [threading.Thread(target=process, args=(i,)) for i in range(n)]

    each("start", threads)
    each("join", threads)

    return payload


def each(accept, iterable, *args, **kwargs):
    """Applies the accept function to each of the elements in the iterable
    collection."""
    if isinstance(accept, str):
        methods = [getattr(item, accept) for item in iterable]
        _ = [method(*args, **kwargs) for method in methods]
        return

    unpack = kwargs.get("_unpack", False)
    for item in iterable:
        _args = (item,) if not unpack else item
        accept(*_args)



"""Exceptions module."""

from http import HTTPStatus


class ConstellationException(Exception):
    """Root class of all exceptions thrown by the constellation client."""

    def __init__(self, message, status_code):
        super().__init__()
        self.message = message
        self.status_code = status_code

    def __str__(self):
        return "{} (StatusCode: {})".format(self.message, self.status_code)


class ConstellationInvalidResponseException(ConstellationException):
    """Thrown when constellation gives an unexpected response."""

    def __init__(self, description, response):
        super().__init__(
            message="{}.  Constellation response: {}".format(description, response),
            status_code=HTTPStatus.INTERNAL_SERVER_ERROR,
        )


class ConstellationInvalidMethodException(ConstellationException):
    """Thrown when attempting an invalid http method for the call to
    constellation."""

    def __init__(self, method, endpoint):
        super().__init__(
            message=" ".join(
                [
                    f"The method {method} is not support for the endpoint {endpoint}.",
                    "Please update your call and try again",
                ]
            ),
            status_code=HTTPStatus.INTERNAL_SERVER_ERROR,
        )


class ConstellationMissingFieldException(ConstellationException):
    """Thrown when a required field is missing."""

    def __init__(self, description, field_name):
        super().__init__(
            message="Missing field: {} when {}".format(field_name, description),
            status_code=HTTPStatus.INTERNAL_SERVER_ERROR,
        )


class ConstellationNotLoggedInException(ConstellationException):
    """Thrown when the user has not logged in and an auth token is required to
    perform the requested action."""

    def __init__(self):
        super().__init__(
            message=" ".join(
                [
                    "You are not logged in.",
                    "Please call login() or set_auth_token() and try again.",
                ]
            ),
            status_code=HTTPStatus.UNAUTHORIZED,
        )


class ConstellationObjectDoesNotExistException(ConstellationException):
    """Thrown when constellation attemping to download an object that does not
    exist."""

    def __init__(self, object_id):
        super().__init__(
            message=" ".join(
                [
                    f"The object {object_id} does not exist or is not accessible by you.",
                    "Please remove it and retry your request.",
                ]
            ),
            status_code=HTTPStatus.NOT_FOUND,
        )


class VariableUnitValueUnparsableExpressionException(BaseException):
    """Thrown when the supplied expression cannot be parsed."""

    def __init__(self, message, expression):
        self._message = message
        self._expression = expression

    def __str__(self):
        return f"VariableUnitValueUnparsableExpressionException: {self._message} for expression: {self._expression}"


class UnknownTokenError(BaseException):
    """Thrown when an unknown token type is encountered.

    Fixing this will require an update to the parser to handle the type
    """

    def __init__(self, token_char, token):
        self._token_char = token_char
        self._token = token

    def __str__(self):
        return f"BinarnyObjectRepresentationUnknownTokenType: {self._token_char} for token {self._token}"


class TraversalTooDeepError(Exception):
    """Thrown when we seemingly lose track of results traversals."""



import os

import pytest

from pysll.decoders import decode


@pytest.fixture(params=["on", "off"])
def _use_numpy(request):
    os.environ["PYSLL_USE_NUMPY"] = request.param
    yield
    del os.environ["PYSLL_USE_NUMPY"]


@pytest.mark.parametrize(
    "encoded,decoded",
    [
        # fmt:off
        ("1:eJxTTMoPymQAAgALRAGu", "0"),
        ("1:eJxTTMoPymRhYGAAAAtUAbI=", "4"),
        ("1:eJxTTMoPytRiYGAAAAvsAdg=", "42"),
        ("1:eJxTTMoP8pRlYGCwNDA2sTC1NLAAUpYGpiAukLIwMLawNDEFAJVXB6k=", "90348590834890590349058038945"),
        ("1:eJxTTMoPKmIAAwEHABKtAgc=", "4.0"),
        ("1:eJxTTMoPKlLeX9fBW8btAAAgJgRt", "3.432643"),
        ("1:eJxTTMoPCuZkYGDwyFQIyUgtSlUEACgoBIs=", "Hi There!"),
        ("1:eJxTTMoPSmMAgmIWIOGTWVwCACIoA74=", "List[]"),
        ("1:eJxTTMoPKmZlYGDwyU/PzAMAGegDtg==", "Login"),
        ("1:eJxTTMoPSmMAgmIWIOGTWVwCACIoA74=", "List[]"),
        ("1:eJxTTMoPSmNiYGAoZgESPpnFJZmMQEYmSAgATfoElQ==", "List[1, 2]"),
        ("1:eJxTTMoPCoplYGAwNDLWA2IgMjE2MTQyAhEIDEFwHrI4ggljJZgb6xlYGhgbW5oZmRgbGpgYAQAiwxSF", "123.12312343412233"),
        (
            "1:eJxNwakBgDAQBMC0Qgeb3O0Tj8FSAg3Qv8MyczzvfZ1jjEDMNKexu+LQU4JKoRplaZsrm2A3klp2VuPnA2uVD6k=",
            "80658175170943878571660636856403766975289505440883277824000000000000"
        ),
        ("1:eJxTTMoPCmZlYGAw1jM0SQAAFg0Cww==", "3.14`"),
        ("1:eJxTTMoPCmZnYGAw1jM0STA0AAAcOwMm", "3.14`10"),
        ("1:eJxTTMoPSmNkYGAoZgESHvk5KRAeJ5BwS0wuyS/KTMzJNAHyALLoCL8=", "Hold[Factorial[52]]"),
        ("1:eJxTTMoPCmZlYGDISM3JyQcAGSIDsQ==", "hello"),
        ("1:eJxTTMoPKmZlYGDISM3JyQcAGmID0Q==", "hello"),  # sus, this is the symbol...
        ("1:eJxTTMoPKmZhYGDwSE1MAQAV6wMu", "Head"),
        ("1:eJxTTMoPCmZhYGDwSE1MAQAUywMO", "Head"),
        ("1:eJxTTMoPSmNmYGAoZgESPpnFJZmMQEYmE4gAiQMAZxgFAg==", "List[1, 2, 3]"),
        ("1:eJxTTMoPSmNkYGAoZgESPpnFJWnMyLxMkFwmE5Aokm99HbhDjtMBAPuoCoE=", "List[List[1, 2, 3.14]]"),
        (
            "1:eJxTTMoPSmNhYGAo5gYSwQWJRcWpjkVFiZXFnEC+Y2lJfm5iSWZyGhNICUidT2ZxiScjkGEMITOBBEMaM7J0JkgCVQfEChT9BmDSCMksE1RVaYwYegyxiBljETPCqg7TDYYQe5FdAgCqsisv",  # noqa: E501
            "SparseArray[Automatic, List[3, 3], 0, List[1, List[List[0, 2, 3, 4], List[List[1], List[3], List[2], List[3]]], List[1, 4, 2, 3]]]"  # noqa: E501
        ),
        ("1:eJxTTMoPSmNkYGAo5gYSjsXF+cmZiSWZ+XlpTCBBFiARVJqTWswMZKTl54PppMQiAF7SDIM=", "Association[Rule[foo, bar]]"),
        ("1:eJxTTMoPCmJnYGAw0jNMMDTSAwAcSwMg", "2.1"),
        (
            "1:eJxTTMoPSmNkYGAo5gUSgaWJeSWZJZWORUWJlWlMIGE5IBFcUlSaXFJalJoClkhA8F0SSxIh6liAhE9mcYknyDAjMGmcxowsg6oOVQ6s3hC3XrCYCZg0BZNmwewgh6Um5+elFKOZzIjMywTx0IRAygHNhzCH",  # noqa: E501
            "QuantityArray[StructuredArray`StructuredData[List[2, 3], List[List[List[1, 2, 3], List[4, 5, 6]], Seconds, List[List[1], List[2]]]]]"  # noqa: E501
        )
        # fmt:on
    ],
)
def test_decode(_use_numpy, encoded, decoded):
    assert decode(encoded) == decoded


def test_large_matrix_decoder():
    """Large matrices of real numbers get compresed into a special format,
    which we test here."""
    with open("tests/decompress-large-matrix.dat", "r") as handle:
        decoded = decode(handle.read())

    from pysll.models import VariableUnitValue

    parsed = VariableUnitValue.parse_mathematica_expression(decoded)

    assert len(parsed) == 361

    first, last = parsed[0], parsed[-1]

    x, y = first
    assert (x.value, x.unit) == (0.0, "Minutes")
    assert (y.value, y.unit) == (-87.528984, "IndependentUnit[Lsus]")

    x, y = last
    assert (x.value, x.unit) == (6.0, "Minutes")
    assert (y.value, y.unit) == (1009.098328, "IndependentUnit[Lsus]")
